/* $Id: ledmodule.c,v 1.5 2012-03-10 13:08:12 simimeie Exp $
 * Functions for communicating with the LED module.
 * The module is connected through an i2c bus. In our implementation, it
 * consists of three SAA 1064 driver ics. There are four possible addresses
 * for the driver ICs, so another driver would be possible without adding
 * an extra bus. This module is not prepared to handle more than one bus.
 */

#ifdef LEDMODULE

#include <avr/io.h>
#include <util/delay.h>
#include "ledmodule.h"
#include "console.h"

/* The Port used for the connection */
#define LEDPORT PORTC
#define LEDPIN PINC
#define LEDDDR DDRC

/* Which pins of the port */
#define SDAPIN PC1
#define SCLPIN PC0

/* the I2C addresses of the SAA 1064 LED drivers */
#define SAA_AD1 0x70   /* Address pin connected to GND */
#define SAA_AD2 0x76   /* Address pin connected to VCC */
#define SAA_AD3 0x72   /* Address pin connected to 3/8 VCC */
#define SAA_AD4 0x74   /* Address pin connected to 5/8 VCC */

#define I2C_READ  0x01
#define I2C_WRITE 0x00

/* Should be at least 27 (80 / 3) at 8 MHz */
/* This was the conservative value used for testing. However, half as much
 * should work as well. */
#define DELAYVAL 40

uint8_t led_brightness;   /* Brightness setting for the LEDs (0-7) */
uint8_t led_dispovon;   /* display override activation */
uint8_t led_dispovbytes[6]; /* the byte values sent to the LED display
                             * for the display override */
uint16_t led_dispovendticks;  /* at which tick counter we stop */

void led_init(void) {
  /* activate pullups */
  LEDPORT |= _BV(SCLPIN);
  LEDPORT |= _BV(SDAPIN);
}

/* For reference:
 * The mapping of the LED panel.
 *   Bit  Wert  Segment
 *    4   0x10     e       *  Segments:
 *    5   0x20     d       *    aaaa
 *    6   0x40     c       *   f    b
 *    7   0x80     P       *   f    b
 *    3   0x08     b       *    gggg
 *    2   0x04     a       *   e    c
 *    1   0x02     f       *   e    c
 *    0   0x01     g       *    dddd  P
 */
uint8_t led_getfontentry(uint8_t c) {
  switch (c) {
  case '0':
  case   0: return 0x7e;
  case '1':
  case   1: return 0x48;
  case '2':
  case   2: return 0x3d;
  case '3':
  case   3: return 0x6d;
  case '4':
  case   4: return 0x4b;
  case '5':
  case   5: return 0x67;
  case '6':
  case   6: return 0x77;
  case '7':
  case   7: return 0x4c;
  case '8':
  case   8: return 0x7f;
  case '9':
  case   9: return 0x6f;
  case 'a':
  case 'A':
  case  10: return 0x5f; /* A */
  case 'b':
  case 'B':
  case  11: return 0x73; /* B */
  case 'c':
  case 'C':
  case  12: return 0x36; /* C */
  case 'd':
  case 'D':
  case  13: return 0x79; /* D */
  case 'e':
  case 'E':
  case  14: return 0x37; /* E */
  case 'f':
  case 'F':
  case  15: return 0x17; /* F */
  case '.': return 0x80;
  case '-': return 0x01;
  default:  return 0xff;
  };
  return 0xff; /* Never reached */
}

/* Send START, defined as high-to-low SDA with SCL high.
 * Expects SCL and SDA to be high already!
 * Returns with SDA and SCL low. */
static void I2C_start(void) {
  /* Change to output mode. */
  LEDDDR |= _BV(SDAPIN);
  LEDDDR |= _BV(SCLPIN);
  /* change SDA to low */
  LEDPORT &= (uint8_t)~_BV(SDAPIN);
  _delay_loop_1(DELAYVAL);
  /* and SCL too */
  LEDPORT &= (uint8_t)~_BV(SCLPIN);
  _delay_loop_1(DELAYVAL);
}

/* Send STOP, defined as low-to-high SDA with SCL high.
 * Expects SCL and SDA to be low already!
 * Returns with SDA and SCL high. */
static void I2C_stop(void) {
  /* Set SCL */
  LEDPORT |= _BV(SCLPIN);
  _delay_loop_1(DELAYVAL);
  /* Set SDA */
  LEDPORT |= _BV(SDAPIN);
  _delay_loop_1(DELAYVAL);
  /* Probably safer to tristate the bus */
  LEDDDR &= (uint8_t)~_BV(SDAPIN);
  LEDDDR &= (uint8_t)~_BV(SCLPIN);
}

/* Transmits the byte in what.
 * Returns 1 if the byte was ACKed, 0 if not.
 * Expects SCL and SDA to be low already!
 * Returns with SDA and SCL low. */
static uint8_t I2C_transmit_byte(uint8_t what) {
  uint8_t i;
  for (i = 0; i < 8; i++) {
    /* First put data on the bus */
    if (what & 0x80) {
      LEDPORT |= _BV(SDAPIN);
    }
    _delay_loop_1(DELAYVAL);
    /* Then set SCL high */
    LEDPORT |= _BV(SCLPIN);
    _delay_loop_1(DELAYVAL);
    /* Take SCL back */
    LEDPORT &= (uint8_t)~_BV(SCLPIN);
    _delay_loop_1(DELAYVAL);
    /* And SDA too */
    LEDPORT &= (uint8_t)~_BV(SDAPIN);
    _delay_loop_1(DELAYVAL);
    what <<= 1;
  }
  /* OK that was the data, now we read back the ACK */
  /* We need to tristate SDA for that */
  LEDPORT |= _BV(SDAPIN);
  LEDDDR &= (uint8_t)~_BV(SDAPIN);
  /* Give the device some time */
  _delay_loop_1(DELAYVAL);
  _delay_loop_1(DELAYVAL);
  _delay_loop_1(DELAYVAL);
  /* Then set SCL high */
  LEDPORT |= _BV(SCLPIN);
  _delay_loop_1(DELAYVAL);
  _delay_loop_1(DELAYVAL);
  _delay_loop_1(DELAYVAL);
  i = LEDPIN & _BV(SDAPIN); /* Read ACK */
#if 0
  if (i) {
    console_printpgm_P(PSTR("!"));
  } else {
    console_printpgm_P(PSTR("A"));
  }
#endif
  /* Take SCL back */
  LEDPORT &= (uint8_t)~_BV(SCLPIN);
  _delay_loop_1(DELAYVAL);
  /* No more tristate, we pull SDA again */
  LEDPORT &= (uint8_t)~_BV(SDAPIN);
  LEDDDR |= _BV(SDAPIN);
  _delay_loop_1(DELAYVAL);
  return (i == 0);
}

void led_set_digit(uint8_t digit, uint8_t val) {
  uint8_t toadr;
  I2C_start();
  /* Address device */
  if (digit < 2) {
    toadr = (SAA_AD1 | I2C_WRITE);
  } else if (digit < 4) {
    toadr = (SAA_AD2 | I2C_WRITE);
  } else if (digit < 6) {
    toadr = (SAA_AD3 | I2C_WRITE);
  } else {
    toadr = (SAA_AD4 | I2C_WRITE);
  }
  I2C_transmit_byte(toadr);
  I2C_transmit_byte((digit & 1) + 1); /* Address Digit Register on device */
  I2C_transmit_byte(val); /* Send value for Digit */
  I2C_stop();
}

void led_set_brightness(void) {
  I2C_start();
  I2C_transmit_byte(SAA_AD1 | I2C_WRITE); /* Address first driver */
  I2C_transmit_byte(0); /* Address Config Register on device */
  I2C_transmit_byte(((led_brightness & 0x07) << 4) | 0x06); /* Send Settings */
  I2C_stop();
  I2C_start();
  I2C_transmit_byte(SAA_AD2 | I2C_WRITE); /* Address second driver */
  I2C_transmit_byte(0); /* Address Config Register on device */
  I2C_transmit_byte(((led_brightness & 0x07) << 4) | 0x06); /* Send Settings */
  I2C_stop();
  I2C_start();
  I2C_transmit_byte(SAA_AD3 | I2C_WRITE); /* Address third driver */
  I2C_transmit_byte(0); /* Address Config Register on device */
  I2C_transmit_byte(((led_brightness & 0x07) << 4) | 0x06); /* Send Settings */
  I2C_stop();
  I2C_start();
  I2C_transmit_byte(SAA_AD4 | I2C_WRITE); /* Address fourth driver */
  I2C_transmit_byte(0); /* Address Config Register on device */
  I2C_transmit_byte(((led_brightness & 0x07) << 4) | 0x06); /* Send Settings */
  I2C_stop();
}

uint8_t led_map(uint8_t n) {
  switch (n) {
  case 0: return 1;
  case 1: return 0;
  case 2: return 3;
  case 3: return 2;
  case 4: return 7;
  case 5: return 6;
  };
  return 8; /* invalid */
}

#endif  /* LEDMODULE */
