/* $Id: main.c,v 1.14 2012-03-10 13:09:09 simimeie Exp $
 * This is software that allows the "universal mainboard" to be used
 * for various things.
 * This is the main file that glues it all together. It also contains all
 * functionality that is too small to require an extra file.
 * (C) Michael "Fox" Meier 2010
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/power.h>
#include <avr/sleep.h>
#include <stdio.h>
#include <string.h>
#include <util/delay.h>

#include "console.h"
#include "timers.h"
#include "network.h"
#include "ledmodule.h"
#include "gps.h"
#include "timeconvert.h"
#include "eepromdata.h"

struct brokentime mbt;

int main(void)
{
  uint8_t i; uint8_t j = 0;
  
  /* Load settings from EEPROM */
  for (i = 0; i < 6; i++) {
    net_mac[i] = eeprom_read_byte(&ee_mac[i]);
  }
  net_fdx = eeprom_read_byte(&ee_fdx);
  for (i = 0; i < 4; i++) {
    net_ip4[i] = eeprom_read_byte(&ee_ip4[i]);
  }
  for (i = 0; i < 16; i++) {
    net_ip6[i] = eeprom_read_byte(&ee_ip6[i]);
  }
  for (i = 0; i < 16; i++) {
    net_rconpass[i] = eeprom_read_byte(&ee_rconpass[i]);
  }
  net_ipv6addrmode = eeprom_read_byte(&ee_ipv6addrmode);
#ifdef LEDMODULE
  led_brightness = eeprom_read_byte(&ee_ledbri);
#endif /* LEDMODULE */
  tc_timezoneoffset = eeprom_read_dword(&ee_timezone);
  tc_handledst = eeprom_read_byte(&ee_handledst);

  /* Initialize stuff */
  console_init();
  timers_init();
  
  led_init();
  led_set_brightness();
  for (i = 0; i < 8; i++) {
    led_set_digit(i, 0xff);
  }
  
  enc28j60_init();
  gps_init();
  
  /* Enable watchdog timer with a timeout of 2 seconds */
  wdt_enable(WDTO_2S);

  /* Disable unused chip functions to save power. Though in this board
   * the power consumption of the AVR is absolutely negligible compared
   * to what the ethernet controller burns anyways, it doesn't hurt to
   * do it. */
  /* Disable timer0, timer2, TWI and the ADC */
  PRR |= _BV(PRTIM2) | _BV(PRTIM0) | _BV(PRADC) | _BV(PRTWI);
  /* Prepare sleep mode */
  set_sleep_mode(SLEEP_MODE_IDLE);
  sleep_enable();

  /* All set up, enable interrupts and go. */
  sei();
  
  if (1) { /* Configure GPS before entering main loop. */
    /* This is not done in the normal init routines because the GPS module
     * needs some time to 'boot' */
    uint8_t gpscfgatt = 0; /* GPS configuration attempts */
    while (gpscfgatt < 20) { /* Attempt at most 20 times */
      wdt_reset();
      if (gps_configurereceiver()) {
        console_printpgm_P(PSTR(" NOTE: GPS ACKnowledged command. "));
        break;
      }
      gpscfgatt++;
    }
  }
  while (1) { /* Main loop, we should never exit it. */
    wdt_reset();
    net_worker();
    if (j == 0) {
      /* Set brightness */
      led_set_brightness();
      if (led_dispovon) { /* Display override active */
        uint16_t dispovdif = led_dispovendticks - getticks();
        if (dispovdif > 0x8000) { /* so the value was essentially negative */
          led_dispovon = 0;
        }
        led_set_digit(led_map(0), led_dispovbytes[0]);
        led_set_digit(led_map(1), led_dispovbytes[1]);
        led_set_digit(led_map(2), led_dispovbytes[2]);
        led_set_digit(led_map(3), led_dispovbytes[3]);
        led_set_digit(led_map(4), led_dispovbytes[4]);
        led_set_digit(led_map(5), led_dispovbytes[5]);
      } else { /* No display override */
        /* Update time display */
        uint8_t mygpsv = gpsvalid;
        if (gpstime[mygpsv].unixts != 0) { /* valid time */
          struct timestamp dts; /* device timestamp */
          uint32_t tsdif; uint32_t tstmp;
          gettimestamp(&dts);
          tsdif = ((uint32_t)dts.tovfls << 16) | dts.ticks;
          tsdif -= ((uint32_t)gpstime[mygpsv].devicetime.tovfls << 16)
                   | gpstime[mygpsv].devicetime.ticks;
          tsdif = (tsdif * 65536UL) / CPUFREQ;
          /* tsdif now contains the number of seconds that have passed since
           * the last signal was received. */
          tstmp = tsdif + gpstime[mygpsv].unixts;
          tstmp += tc_timezoneoffset;
          if ((tc_handledst) && (gpstime[mygpsv].issummertime)) {
            tstmp += 3600;
          }
          tstobt(tstmp, &mbt);
          if ((32 < (mbt.sec % 60)) && ((mbt.sec % 60) < 38)) { /* display date */
            led_set_digit(led_map(0), led_getfontentry(mbt.day / 10));
            led_set_digit(led_map(1), led_getfontentry(mbt.day % 10) | led_getfontentry('.'));
            led_set_digit(led_map(2), led_getfontentry(mbt.month / 10));
            led_set_digit(led_map(3), led_getfontentry(mbt.month % 10) | led_getfontentry('.'));
            led_set_digit(led_map(4), led_getfontentry(mbt.year / 10));
            led_set_digit(led_map(5), led_getfontentry(mbt.year % 10));
          } else { /* display time */
            led_set_digit(led_map(0), led_getfontentry(mbt.hour / 10));
            led_set_digit(led_map(1), led_getfontentry(mbt.hour % 10));
            led_set_digit(led_map(2), led_getfontentry(mbt.min / 10));
            led_set_digit(led_map(3), led_getfontentry(mbt.min % 10));
            led_set_digit(led_map(4), led_getfontentry(mbt.sec / 10));
            led_set_digit(led_map(5), led_getfontentry(mbt.sec % 10));
          }
          if (tsdif > 300) { /* 5 minutes without a signal? useless then. */
            gpstime[mygpsv].unixts = 0; /* mark invalid. */
          }
        } else { /* No valid time. display something? */
          uint8_t kniri = (getticks() / (TICKSPERSECOND / 4)) % 11;
          if (kniri >= 6) {
            kniri = 11 - kniri;
          }
          for (i = 0; i < 6; i++) {
            uint8_t whattoshow = 0;
            if (kniri == i) {
              whattoshow = led_getfontentry('-');
            }
            if (gpssatsinview > i) {
              whattoshow |= led_getfontentry('.');
            }
            led_set_digit(led_map(i), whattoshow);
          }
        }
      }
    }
    j = (j + 1) & 0x0f; /* i.e. modulo 16 */
    sleep_cpu(); /* Go to sleep until the next IRQ arrives */
  }
}
